#!/usr/bin/env python3
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import codecs
import os
import re


def read(*parts):
    """
    Build an absolute path from ``parts`` and return contents of the
    resulting file. Assumes UTF-8 encoding.
    """
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, *parts), 'rb', 'utf-8') as f:
        return f.read()


def find_version(*file_paths):
    """
    Build a path from ``file_paths`` and search for ``__version__``.
    """
    version_file = read(*file_paths)
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)

    if version_match:
        return version_match.group(1)
    raise RuntimeError('Unable to find version string.')


# -- General configuration ------------------------------------------------

# Sphinx extension modules
extensions = [
    # include documentation from docstrings
    'sphinx.ext.autodoc',
    # highlight todo directives in doc compilation
    'sphinx.ext.todo',
    # docstring coverage
    'sphinx.ext.coverage',
    # link to view rST code
    'sphinx.ext.viewcode',
    # Google-style docstring formatting
    'sphinx.ext.napoleon',
    # link to other projects' documentation
    'sphinx.ext.intersphinx',
    # allow autodoc to work with zope.interface interfaces
    'repoze.sphinx.autointerface',
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'gordon-dns'
copyright = '2017, Spotify AB'
author = 'Matt Obarzanek, Lynn Root, Amy Skinner, Steven Stone, James Wen'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The full version, including alpha/beta/rc tags.
release = find_version('../gordon/__init__.py')
# The short X.Y version.
version = release.rsplit(u'.', 1)[0]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = ['_build']

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True


# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = '_static/spotify_icon_green.png'

# The name of an image file (within the static path) to use as favicon of the
# docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = '_static/spotify_icon_black.ico'

# If false, no module index is generated.
html_domain_indices = True

# If false, no index is generated.
html_use_index = True

# If true, the index is split into individual pages for each letter.
html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'gordondnsdoc'


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    (master_doc, 'gordon-dns', u'gordon Documentation',
     [author], 1)
]


# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {'https://docs.python.org/3': None}



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pkg_resources
import pytest
import zope.interface

from gordon import exceptions
from gordon import interfaces
from gordon import plugins_loader
from gordon.metrics import ffwd
from tests.unit import conftest


#####
# Fixtures
#####
@pytest.fixture(scope='session')
def namespaced_config():
    return {
        'event_consumer': {'a_key': 'a_value', 'b_key': 'b_value'},
        'event_consumer.plugin': {'a_key': 'another_value'},
        'enricher': {},
        'enricher.plugin': {'d_key': 'd_value'}
    }


@pytest.fixture(scope='session')
def plugin_config():
    return {
        'xyz': {
            'a_key': 'a_value',
            'b_key': 'b_value',
        },
        'xyz.event_consumer': {
            'a_key': 'another_value',
            'b_key': 'b_value'
        },
        'xyz.enricher': {
            'a_key': 'a_value',
            'b_key': 'b_value',
            'd_key': 'd_value',
        },
        'xyz.publisher': {
            'a_key': 'a_value',
            'b_key': 'b_value',
            'c_key': 'c_value',
        }
    }


@pytest.fixture
def exp_inited_plugins(plugin_config, plugin_kwargs):
    return [
        conftest.EventConsumerStub(
            plugin_config['xyz.event_consumer'], **plugin_kwargs),
        conftest.EnricherStub(
            plugin_config['xyz.enricher'], **plugin_kwargs),
        conftest.PublisherStub(
            plugin_config['xyz.publisher'], **plugin_kwargs)
    ]


@pytest.fixture
def mock_iter_entry_points(mocker, monkeypatch, installed_plugins):
    mock_plugins = installed_plugins.values()

    mock_iter_entry_points = mocker.MagicMock(pkg_resources.iter_entry_points)
    mock_iter_entry_points.return_value = iter(mock_plugins)
    monkeypatch.setattr(plugins_loader.pkg_resources, 'iter_entry_points',
                        mock_iter_entry_points)


def is_instance_of_stub(obj):
    stubs = [
        conftest.EventConsumerStub,
        conftest.EnricherStub,
        conftest.PublisherStub,
        conftest.GenericStub
    ]
    return any([isinstance(obj, stub) for stub in stubs])


#####
# The good stuff
#####
def test_init_plugins(installed_plugins, plugin_config, inited_plugins,
                      plugin_kwargs):
    """Plugins are initialized with their config."""
    inited_names, inited_plugins, errors = plugins_loader._init_plugins(
        conftest.REGISTERED_ACTIVE_PLUGINS, installed_plugins, plugin_config,
        plugin_kwargs
    )

    assert sorted(conftest.REGISTERED_ACTIVE_PLUGINS) == sorted(inited_names)
    for plugin_obj in inited_plugins:
        assert is_instance_of_stub(plugin_obj)
        assert any([p.config == plugin_obj.config for p in inited_plugins])


def test_init_plugins_exceptions(mocker, plugin_kwargs):
    """Non-callable plugin returns plugin-specific exceptions."""
    name = 'B0rkedPlugin'
    config = {'B0rkedPlugin': {'foo': 'bar'}}

    plugin_mock = mocker.MagicMock(pkg_resources.EntryPoint, autospec=True)
    plugin_mock.name = name
    plugin_mock.load.return_value = 'not_a_class'
    plugins = {name: plugin_mock}

    inited_names, inited_plugins, errors = plugins_loader._init_plugins(
        [name], plugins, config, plugin_kwargs)
    assert 1 == len(errors)


def test_init_plugins_skipped(installed_plugins, plugin_config, caplog,
                              plugin_kwargs):
    """Skips plugins that are not configured."""
    config = {'xyz.event_consumer': plugin_config['xyz.event_consumer']}

    inited_names, inited_plugins, errors = plugins_loader._init_plugins(
        conftest.REGISTERED_ACTIVE_PLUGINS, installed_plugins, config,
        plugin_kwargs
    )

    assert 1 == len(inited_plugins) == len(inited_names)
    assert 2 == len(caplog.records)


def test_init_plugins_empty_config(installed_plugins, plugin_kwargs):
    """Loads plugin if mathcing config key exists with empty config."""
    config = {name: {} for name in conftest.REGISTERED_ACTIVE_PLUGINS}

    inited_names, inited_plugins, errors = plugins_loader._init_plugins(
        conftest.REGISTERED_ACTIVE_PLUGINS, installed_plugins, config,
        plugin_kwargs
    )

    assert 3 == len(inited_plugins) == len(inited_names)
    for plugin_obj in inited_plugins:
        assert {} == plugin_obj.config


def test_init_plugins_skip_inactive(installed_plugins, plugin_config,
                                    plugin_kwargs):
    """Skips plugins that are not activated in core config."""
    inited_names, inited_plugins, errors = plugins_loader._init_plugins(
        [conftest.REGISTERED_ACTIVE_PLUGINS[0]], installed_plugins,
        plugin_config, plugin_kwargs)

    assert 1 == len(inited_plugins) == len(inited_names)
    exp = plugin_config.get(conftest.REGISTERED_ACTIVE_PLUGINS[0])
    assert exp == inited_plugins[0].config


@pytest.mark.parametrize('namespace,exp_config', (
    ('event_consumer', {'a_key': 'a_value', 'b_key': 'b_value'}),
    ('event_consumer.plugin', {'a_key': 'another_value', 'b_key': 'b_value'}),
    ('enricher', {}),
    ('enricher.plugin', {'d_key': 'd_value'})
))
def test_merge_config(namespace, exp_config, namespaced_config):
    """Namespaced config for a plugin also has parent/global config."""
    ret_config = plugins_loader._merge_config(namespaced_config, namespace)

    assert exp_config == ret_config


@pytest.mark.parametrize('namespace,exp_config', (
    ('xyz', {'a_key': 'a_value', 'b_key': 'b_value'}),
    ('xyz.event_consumer', {'a_key': 'another_value'}),
    ('xyz.enricher', {'d_key': 'd_value'}),
))
def test_get_namespaced_config(namespace, exp_config, installed_plugins,
                               loaded_config):
    """Tease out config specific to a plugin with no parent config."""
    all_plugins = installed_plugins.keys()
    ret_namespace, ret_config = plugins_loader._get_namespaced_config(
        loaded_config, namespace, all_plugins)

    assert exp_config == ret_config
    assert namespace == ret_namespace


def test_load_plugin_configs(installed_plugins, loaded_config, plugin_config):
    """Load plugin-specific config ignoring other plugins' configs."""
    plugin_names = ['xyz'] + conftest.REGISTERED_ACTIVE_PLUGINS
    parsed_config = plugins_loader._load_plugin_configs(
        plugin_names, loaded_config)

    for name in conftest.REGISTERED_ACTIVE_PLUGINS:
        assert plugin_config[name] == parsed_config[name]


def test_get_plugin_config_keys(installed_plugins):
    """Entry point keys for plugins are parsed to config keys."""
    config_keys = plugins_loader._get_plugin_config_keys(installed_plugins)
    expected = ['xyz'] + conftest.REGISTERED_PLUGINS
    assert sorted(expected) == sorted(config_keys)


def test_get_activated_plugins(loaded_config, installed_plugins):
    """Assert activated plugins are installed."""
    active = plugins_loader._get_activated_plugins(
        loaded_config, installed_plugins)

    assert conftest.REGISTERED_ACTIVE_PLUGINS == active


def test_get_activated_plugins_raises(loaded_config, installed_plugins):
    """Raise when activated plugins are not installed."""
    loaded_config['core']['plugins'].append('xyz.not_installed_plugin')

    with pytest.raises(exceptions.LoadPluginError) as e:
        plugins_loader._get_activated_plugins(loaded_config, installed_plugins)

    e.match('Plugin "xyz.not_installed_plugin" not installed')


def test_gather_installed_plugins(mock_iter_entry_points, installed_plugins):
    """Gather entry points/plugins into a {name: entry point} format."""
    gathered_plugins = plugins_loader._gather_installed_plugins()
    assert sorted(installed_plugins) == sorted(gathered_plugins)


def test_load_plugins(mock_iter_entry_points, loaded_config, installed_plugins,
                      exp_inited_plugins, plugin_kwargs):
    """Plugins are loaded and instantiated with their config."""
    inited_names, installed_plugins, errors, _ = plugins_loader.load_plugins(
        loaded_config, plugin_kwargs)

    assert 3 == len(inited_names) == len(installed_plugins)
    for plugin_obj in installed_plugins:
        assert is_instance_of_stub(plugin_obj)
        assert any([p.config == plugin_obj.config for p in exp_inited_plugins])


def test_load_plugins_none_loaded(mocker, installed_plugins, plugin_kwargs):
    """Return empty list when no plugins are found."""
    mock_iter_entry_points = mocker.MagicMock(pkg_resources.iter_entry_points)
    mock_iter_entry_points.return_value = []

    loaded_config = {'core': {}}

    inited_names, installed_plugins, errors, _ = plugins_loader.load_plugins(
        loaded_config, plugin_kwargs)
    assert [] == installed_plugins == inited_names == errors


def test_load_plugins_exceptions(installed_plugins, loaded_config,
                                 mock_iter_entry_points, plugin_exc_mock,
                                 plugin_kwargs, mocker, monkeypatch):
    """Loading plugin exceptions are returned."""
    inited_plugins_mock = mocker.MagicMock(
        plugins_loader._init_plugins, autospec=True)

    exc = [('bad.plugin', plugin_exc_mock)]
    inited_plugins_mock.return_value = (
        conftest.REGISTERED_PLUGINS, inited_plugins_mock, exc)
    monkeypatch.setattr(plugins_loader, '_init_plugins', inited_plugins_mock)

    inited_names, installed_plugins, errors, _ = plugins_loader.load_plugins(
        loaded_config, plugin_kwargs)
    assert 1 == len(errors)


@zope.interface.implementer(interfaces.IMetricRelay)
class MetricRelayStub:
    def __init__(self, config):
        pass


@pytest.fixture
def metrics_mock(mocker):
    relay_mock = mocker.MagicMock(pkg_resources.EntryPoint)
    relay_mock.name = 'mock-provider-name'
    relay_mock.load.return_value = MetricRelayStub
    return relay_mock


@pytest.fixture
def plugins_incl_metrics(mocker, monkeypatch, metrics_mock, installed_plugins):
    installed_plugins[metrics_mock.name] = metrics_mock
    mock_iter_entry_points = mocker.Mock(pkg_resources.iter_entry_points)
    mock_iter_entry_points.return_value = iter(installed_plugins.values())
    monkeypatch.setattr(plugins_loader.pkg_resources, 'iter_entry_points',
                        mock_iter_entry_points)
    return installed_plugins


def test_load_plugins_with_metrics(plugins_incl_metrics, loaded_config,
                                   exp_inited_plugins, plugin_kwargs,
                                   metrics_mock):
    """Plugins are loaded and instantiated with their config and metrics."""
    loaded_config['core'].update({'metrics': metrics_mock.name})
    names, installed_plugins, errors, metrics = plugins_loader.load_plugins(
        loaded_config, plugin_kwargs)

    # if metrics were included, len() would be 4
    assert 3 == len(names) == len(installed_plugins)
    for plugin_obj in installed_plugins:
        assert not isinstance(plugin_obj, MetricRelayStub)
        assert is_instance_of_stub(plugin_obj)
        assert any([p.config == plugin_obj.config for p in exp_inited_plugins])
        assert isinstance(metrics, MetricRelayStub)


def test_get_metrics_returns_ffwd(loaded_config, plugins_incl_metrics):
    loaded_config['core'].update({'metrics': 'ffwd'})
    actual = plugins_loader._get_metrics_plugin(
        loaded_config, plugins_incl_metrics)
    assert isinstance(actual, ffwd.SimpleFfwdRelay)


def test_get_metrics_returns_plugin(metrics_mock, plugins_incl_metrics):
    """MetricRelay should load if both implements interface and configured."""
    config = {'core': {'metrics': 'mock-provider-name'}}
    actual = plugins_loader._get_metrics_plugin(config, plugins_incl_metrics)
    assert isinstance(actual, MetricRelayStub)


def test_get_metrics_not_installed_raises(installed_plugins):
    """Return None if config or name incorrect."""
    config = {'core': {'metrics': 'non-installed-metrics-provider'}}
    with pytest.raises(exceptions.LoadPluginError) as e:
        plugins_loader._get_metrics_plugin(config, installed_plugins)

    assert e.match('Metrics.*non-installed-metrics-provider.*not installed')



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module for reusable pytest fixtures.
"""

import asyncio
import logging
import os

import pkg_resources
import pytest
import zope.interface

from gordon import interfaces


PLUGIN_NAMES = [
    'event_consumer',
    'enricher',
    'publisher',
    'runnable'
]
# what plugins are registered as within `setup.py:setup.entry_points`
REGISTERED_PLUGINS = ['xyz.' + p for p in PLUGIN_NAMES]
ACTIVE_NAMES = PLUGIN_NAMES[:3]
REGISTERED_ACTIVE_PLUGINS = REGISTERED_PLUGINS[:3]


@zope.interface.implementer(interfaces.IRunnable, interfaces.IMessageHandler)
class EventConsumerStub:
    start_phase = 'consume'
    phase = 'cleanup'

    def __init__(self, config, success_channel, error_channel, metrics=None,
                 **kwargs):
        self.config = config
        self.success_channel = success_channel
        self.error_channel = error_channel
        self._mock_run_count = 0
        self._mock_handle_message_count = 0

    async def handle_message(self, event_msg):
        await asyncio.sleep(0)
        self._mock_handle_message_count += 1

    async def run(self):
        await asyncio.sleep(0)
        self._mock_run_count += 1

    async def shutdown(self):
        pass


@zope.interface.implementer(interfaces.IMessageHandler)
class EnricherStub:
    phase = 'enrich'

    def __init__(self, config, metrics=None, **kwargs):
        self.config = config
        self._mock_handle_message_count = 0

    async def handle_message(self, event_msg):
        await asyncio.sleep(0)
        self._mock_handle_message_count += 1

    async def shutdown(self):
        pass


class PublisherStub(EnricherStub):
    phase = 'publish'


@zope.interface.implementer(interfaces.IRunnable)
class GenericStub:
    def __init__(self, config, success_channel, error_channel, metrics=None):
        self.config = config
        self.success_channel = success_channel
        self.error_channel = error_channel
        self._mock_run_count = 0

    async def run(self):
        await asyncio.sleep(0)
        self._mock_run_count += 1

    async def shutdown(self):
        pass


@pytest.fixture
def plugin_exc_mock():
    class FakePluginException(Exception):
        """Exception raised from a plugin when loading"""
        pass
    return FakePluginException('dangit')


@pytest.fixture(scope='session')
def config_file():
    here = os.path.dirname(os.path.realpath(__file__))
    filepath = os.path.join(here, 'fixtures/test-gordon.toml')
    with open(filepath, 'r') as f:
        return f.read()


@pytest.fixture
def loaded_config():
    return {
        'core': {
            'plugins': [
                'xyz.event_consumer',
                'xyz.enricher',
                'xyz.publisher'],
            'debug': True,
            'logging': {
                'level': 'debug',
                'handlers': ['stream'],
            },
            'route': {
                'consume': 'enrich',
                'enrich': 'publish',
                'publish': 'cleanup'
            }
        },
        'xyz': {
            'a_key': 'a_value',
            'b_key': 'b_value',
            'event_consumer': {
                'a_key': 'another_value',
            },
            'enricher': {
                'd_key': 'd_value',
            },
            'publisher': {
                'c_key': 'c_value',
            }
        }
    }


def _get_stub_class(plugin_name):
    plugin_name = plugin_name.split('.')[-1]
    return {
        'event_consumer': EventConsumerStub,
        'enricher': EnricherStub,
        'publisher': PublisherStub,
        'runnable': GenericStub
    }[plugin_name]


def _get_plugin_conf(plugin_name):
    conf = loaded_config()
    try:
        return conf[plugin_name]
    except KeyError:
        return {}


@pytest.fixture
def installed_plugins(mocker):
    def mock_plugin(name):
        plugin_mock = mocker.MagicMock(pkg_resources.EntryPoint)
        plugin_mock.name = name
        plugin_mock.load.return_value = _get_stub_class(name)
        return plugin_mock

    installed = {}
    for plugin_name in REGISTERED_PLUGINS:
        installed[plugin_name] = mock_plugin(plugin_name)
    return installed


@pytest.fixture
def caplog(caplog):
    """Set global test logging levels."""
    caplog.set_level(logging.DEBUG)
    logging.getLogger('asyncio').setLevel(logging.WARNING)
    return caplog


@pytest.fixture
def plugin_kwargs(event_loop):
    return {
        'success_channel': asyncio.Queue(),
        'error_channel': asyncio.Queue(),
    }


@pytest.fixture
def inited_plugins(plugin_kwargs):
    plugins = {}
    for name in PLUGIN_NAMES:
        conf = _get_plugin_conf(name)
        plugins[name] = _get_stub_class(name)(conf, **plugin_kwargs)

    return plugins



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module for loading plugins distributed via third-party packages.

Plugin discovery is done via ``entry_points`` defined in a package's
``setup.py``, registered under ``'gordon.plugins'``. For example:

.. code-block:: python

    # setup.py
    from setuptools import setup

    setup(
        name=NAME,
        # snip
        entry_points={
            'gordon.plugins': [
                'gcp.gpubsub = gordon_gcp.gpubsub:EventClient',
                'gcp.gce.a = gordon_gcp.gce.a:ReferenceSourceClient',
                'gcp.gce.b = gordon_gcp.gce.b:ReferenceSourceClient',
                'gcp.gdns = gordon_gcp.gdns:DNSProviderClient',
            ],
        },
        # snip
    )

Plugins are initialized with any config defined in ``gordon-user.toml``
and ``gordon.toml``. See :doc:`config` for more details.

Once a plugin is found, the loader looks up its configuration via the
same key defined in its entry point, e.g. ``gcp.gpubsub``.

The value of the entry point (e.g. ``gordon_gcp.gpubsub:EventClient``)
must point to a class. The plugin class is instantiated with its config.

A plugin will not have access to another plugin's configuration. For
example, the ``gcp.gpusub`` will not have access to the configuration
for ``gcp.gdns``.

See :doc:`plugins` for details on how to write a plugin for Gordon.
"""

import logging

import pkg_resources

from gordon import exceptions
from gordon.metrics import ffwd
from gordon.metrics import log


PLUGIN_NAMESPACE = 'gordon.plugins'


def _init_plugins(active, installed, configs, kwargs):
    inited_plugins, inited_plugin_names, errors = [], [], []
    for plugin_name, unloaded_plugin in installed.items():
        if plugin_name not in active:
            continue
        plugin_class = unloaded_plugin.load()
        plugin_config = configs.get(plugin_name)

        # check against `None` because `plugin_config` could be `{}`,
        # which should be handled by the plugin's logic (i.e. accept all
        # defaults, or raise error, or something else)
        if plugin_config is None:
            msg = (f'Skipped loading plugin "{plugin_name}" because no '
                   'configuration was found.')
            logging.info(msg)
            continue

        try:
            plugin_object = plugin_class(plugin_config, **kwargs)
            inited_plugins.append(plugin_object)
            inited_plugin_names.append(plugin_name)
        except Exception as e:
            errors.append((plugin_name, e))

    return inited_plugin_names, inited_plugins, errors


def _merge_config(config, namespace):
    config_copy = config.copy()
    keys_to_merge = namespace.split('.')

    # DFS approach in order to prefer child config to parent config
    merged_plugin_config = {}
    while keys_to_merge:
        ns = '.'.join(keys_to_merge)
        plugin_config = config_copy.get(ns, {})
        plugin_config_copy = plugin_config.copy()
        plugin_config_copy.update(merged_plugin_config)
        merged_plugin_config = plugin_config_copy
        keys_to_merge.pop()
    return merged_plugin_config


def _get_namespaced_config(config, plugin_namespace, all_plugins):
    plugin_config_keys = plugin_namespace.split('.')

    # drill down to get config that matches plugin namespace
    while plugin_config_keys:
        key = plugin_config_keys.pop(0)
        config = config.get(key)

    # find which config namespaces map to other plugins
    plugin_config_keys_to_exclude = set()
    for plugin in all_plugins:
        if plugin == plugin_namespace:
            continue
        if plugin.startswith(plugin_namespace):
            # exclude same level & lower config keys for other plugins
            keys_to_exclude = plugin.lstrip(plugin_namespace).split('.')
            plugin_config_keys_to_exclude.update(keys_to_exclude)

    # clean up config by removing other plugin configs
    plugin_config = config.copy()
    while plugin_config_keys_to_exclude:
        key = plugin_config_keys_to_exclude.pop()
        try:
            plugin_config.pop(key)
        except KeyError:
            pass

    return plugin_namespace, plugin_config


def _load_plugin_configs(plugin_names, config):
    # A plugin should only have access to its own config and
    # parent/global plugin config, but not configs of other plugins
    plugin_configs = {}
    for plugin in plugin_names:
        namespace, plugin_conf = _get_namespaced_config(
            config, plugin, plugin_names)
        plugin_configs[namespace] = plugin_conf

    # merge namespaced with parent / global plugin config
    merged_configs = {}
    for namespace in plugin_configs.keys():
        plugin_config = _merge_config(plugin_configs, namespace)
        merged_configs[namespace] = plugin_config

    return merged_configs


def _get_plugin_config_keys(plugins):
    # Make sure all parent namespaces of a plugin are available
    # to load config for easy config handling
    all_config_keys = set()
    for namespace in plugins:
        namespaces = namespace.split('.')
        namespaces_to_build = []
        while len(namespaces):
            namespace = namespaces.pop(0)
            namespaces_to_build.append(namespace)
            config_key = '.'.join(namespaces_to_build)
            all_config_keys.add(config_key)
    return sorted(list(all_config_keys))


def _get_activated_plugins(config, installed_plugins):
    activated_plugins = config.get('core', {}).get('plugins', [])
    for active_plugin in activated_plugins:
        if active_plugin not in installed_plugins:
            msg = f'Plugin "{active_plugin}" not installed'
            raise exceptions.LoadPluginError(msg)
    return activated_plugins


def _gather_installed_plugins():
    gathered_plugins = {}
    for entry_point in pkg_resources.iter_entry_points(PLUGIN_NAMESPACE):
        gathered_plugins[entry_point.name] = entry_point
    return gathered_plugins


def _get_metrics_plugin(config, installed_plugins):
    metrics_provider = config.get('core', {}).get('metrics', 'metrics-logger')
    metrics_config = config.get(metrics_provider, {})

    if metrics_provider == 'metrics-logger':
        return log.LogRelay(metrics_config)

    if metrics_provider == 'ffwd':
        return ffwd.SimpleFfwdRelay(metrics_config)

    for plugin_name, plugin in installed_plugins.items():
        if metrics_provider == plugin_name:
            plugin_class = plugin.load()
            return plugin_class(metrics_config)

    msg = f'Metrics Plugin "{metrics_provider}" configured, but not installed'
    raise exceptions.LoadPluginError(msg)


def load_plugins(config, plugin_kwargs):
    """
    Discover and instantiate plugins.

    Args:
        config (dict): loaded configuration for the Gordon service.
        plugin_kwargs (dict): keyword arguments to give to plugins
            during instantiation.
    Returns:
        Tuple of 3 lists: list of names of plugins, list of
        instantiated plugin objects, and any errors encountered while
        loading/instantiating plugins. A tuple of three empty lists is
        returned if there are no plugins found or activated in gordon
        config.
    """
    installed_plugins = _gather_installed_plugins()
    metrics_plugin = _get_metrics_plugin(config, installed_plugins)
    if metrics_plugin:
        plugin_kwargs['metrics'] = metrics_plugin

    active_plugins = _get_activated_plugins(config, installed_plugins)
    if not active_plugins:
        return [], [], [], None
    plugin_namespaces = _get_plugin_config_keys(active_plugins)
    plugin_configs = _load_plugin_configs(plugin_namespaces, config)
    plugin_names, plugins, errors = _init_plugins(
        active_plugins, installed_plugins, plugin_configs, plugin_kwargs)
    return plugin_names, plugins, errors, metrics_plugin



# -*- coding: utf-8 -*-
# Copyright (c) 2018 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers

import pytest

from gordon.metrics import log


@pytest.mark.parametrize('context,context_str', [
    ({'source': 'external'}, ' context: {\'source\': \'external\'}'),
    ({}, '')
])
def test_logger_adapter_log_outputs_message(caplog, context, context_str):
    """Metric is formatted and output to log."""
    logger_adapter = log.LoggerAdapter('INFO')
    metric = {
        'metric_name': 'events',
        'value': 12.12,
        'context': context
        }
    logger_adapter.log(metric)
    assert 1 == len(caplog.records)
    expected_log = '[events] value: 12.12' + context_str
    assert expected_log == caplog.records[0].msg


@pytest.fixture
def mock_logger_adapter(mocker):
    return mocker.Mock(log.LoggerAdapter)


@pytest.fixture
def log_timer(mock_logger_adapter):
    metric = {'metric_name': 'test-metric', 'value': 13.05, 'context': {}}
    return log.LogTimer(metric, mock_logger_adapter, 1)


@pytest.mark.asyncio
async def test_log_timer_manual(log_timer):
    """Times when using start/stop interface."""

    await log_timer.start()
    assert isinstance(log_timer._start_time, numbers.Number)

    await log_timer.stop()
    assert 1 == log_timer.logger.log.call_count
    actual_metric = log_timer.logger.log.mock_calls[0][1][0]
    assert 'test-metric' == actual_metric['metric_name']
    assert {} == actual_metric['context']
    assert 0 < actual_metric['value']


@pytest.mark.asyncio
async def test_log_timer_as_context_manager(log_timer):
    """Times when using as a context manager."""
    async with log_timer as t:
        assert isinstance(t._start_time, numbers.Number)
    actual_metric = log_timer.logger.log.mock_calls[0][1][0]
    assert 1 == log_timer.logger.log.call_count
    assert 'test-metric' == actual_metric['metric_name']
    assert {} == actual_metric['context']
    assert 0 < actual_metric['value']


@pytest.fixture
def relay_config():
    return {
        'time_unit': 1,
        'log_level': 'info',
    }


@pytest.mark.parametrize('name,val,context', [
    ('a-metric', 12, None),
    ('nother-metric', 0.0003, {'source': 'internal'}),
])
def test_logrelay_create_metric(relay_config, name, val, context):
    """Create a metric dictionary."""
    relay = log.LogRelay(relay_config)
    actual = relay._create_metric(name, val, context=context)

    expected = {
        'metric_name': name,
        'value': val,
        'context': context or {}
    }
    assert expected == actual


@pytest.fixture
def log_relay_inst(monkeypatch, mocker, relay_config):
    logger_mock = mocker.Mock(log.LoggerAdapter)
    monkeypatch.setattr(log, 'LoggerAdapter', logger_mock)
    return log.LogRelay(relay_config)


@pytest.mark.asyncio
@pytest.mark.parametrize('context', [
    {'source': 'perimiter'},
    None
])
async def test_logrelay_incr(log_relay_inst, context):
    """Initializes and outputs metrics to logger."""
    expected_metric = log_relay_inst._create_metric('requests-recv', 601,
                                                    context=context)
    await log_relay_inst.incr('requests-recv', 601, context=context)

    log_relay_inst.logger.log.assert_called_once_with(expected_metric)


@pytest.mark.asyncio
async def test_logrelay_incr_accumulation(mocker, log_relay_inst):
    """MetricRelay.incr accumulates values."""
    base_value = 441
    incr = 10
    expected_metric = log_relay_inst._create_metric('requests-recv',
                                                    base_value, context=None)
    expected_2nd_metric = log_relay_inst._create_metric(
        'requests-recv', base_value + incr, context=None)
    await log_relay_inst.incr('requests-recv', base_value)
    await log_relay_inst.incr('requests-recv', incr)

    assert base_value + incr == log_relay_inst.counters['requests-recv']
    expected_calls = [
        mocker.call(expected_metric),
        mocker.call(expected_2nd_metric)
    ]
    log_relay_inst.logger.log.assert_has_calls(expected_calls)


def test_logrelay_timer(log_relay_inst):
    """Initialize and return a LogTimer object."""
    expected_metric = log_relay_inst._create_metric('latency', None,
                                                    context=None)
    timer = log_relay_inst.timer('latency')
    assert isinstance(timer, log.LogTimer)
    assert expected_metric == timer.metric
    assert timer.logger == log_relay_inst.logger


@pytest.mark.asyncio
@pytest.mark.parametrize('context', [
    {'version': 'dev0'},
    None
])
async def test_logrelay_set(log_relay_inst, context):
    """Output single metric to logger."""
    expected_metric = log_relay_inst._create_metric('latency', 451,
                                                    context=context)
    await log_relay_inst.set('latency', 451, context=context)

    log_relay_inst.logger.log.assert_called_once_with(expected_metric)


@pytest.mark.asyncio
async def test_cleanup(log_relay_inst):
    """Implemented, but a noop."""
    assert await log_relay_inst.cleanup() is None



# -*- coding: utf-8 -*-
#
# Copyright 2018 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import zope.interface

from gordon import interfaces
from gordon import router


@zope.interface.implementer(interfaces.IEventMessage)
class EventMsgStub:
    def __init__(self, mocker):
        self.msg_id = 1234
        self.phase = None
        self._mock_update_phase_count = 0
        self._mock_append_to_history = 0

    def update_phase(self, new_phase):
        self.phase = new_phase
        self._mock_update_phase_count += 1

    def append_to_history(self, *args, **kwargs):
        self._mock_append_to_history += 1

    def __repr__(self):
        return f'EventMsgStub(msg_id={self.msg_id})'


@zope.interface.implementer(interfaces.IMetricRelay)
class MetricRelayStub:
    def __init__(self, config):
        self.config = config
        self._mock_incr_call_count = 0
        self._mock_incr_call_args = []

        self._mock_timer_call_count = 0
        self._mock_timer_call_args = []

        self._mock_set_call_count = 0
        self._mock_set_call_args = []

    async def incr(self, metric_name, value=1, context=None, **kwargs):
        self._mock_incr_call_count += 1
        self._mock_incr_call_args.append((metric_name, value, context, kwargs))

    def timer(self, metric_name, context=None, **kwargs):
        self._mock_timer_call_count += 1
        self._mock_timer_call_args.append((metric_name, context, kwargs))
        return TimerStub(metric_name)

    async def set(self, metric_name, value, context=None, **kwargs):
        self._mock_set_call_count += 1
        self._mock_set_call_args.append((metric_name, value, context, kwargs))


@zope.interface.implementer(interfaces.ITimer)
class TimerStub:
    def __init__(self, metric_name):
        self.metric_name = metric_name
        self._mock_start_call_count = 0
        self._mock_stop_call_count = 0

    async def start(self, *args, **kwargs):
        self._mock_start_call_count += 1

    async def stop(self, *args, **kwargs):
        self._mock_stop_call_count += 1


@pytest.fixture
def metrics():
    return MetricRelayStub({})


@pytest.fixture()
def router_inst(loaded_config, inited_plugins, metrics, plugin_kwargs):
    inited_plugins.pop('runnable')
    route_config = loaded_config['core']['route']
    router_inst = router.GordonRouter(
            route_config, plugins=inited_plugins.values(), metrics=metrics,
            **plugin_kwargs)
    return router_inst


@pytest.fixture
def event_msg(mocker):
    msg = EventMsgStub(mocker)
    msg.phase = 'consume'
    return msg


@pytest.mark.parametrize('has_enricher', [True, False])
def test_init_router(has_enricher, inited_plugins, plugin_kwargs, metrics):
    exp_phase_route = {
        'consume': 'publish',
        'publish': 'cleanup',
        'cleanup': 'cleanup'
    }

    exp_phase_map = {
        'consume': interfaces.IRunnable,
        'publish': interfaces.IMessageHandler,
        'cleanup': interfaces.IMessageHandler
    }

    inited_plugins.pop('runnable')
    if has_enricher:
        exp_phase_route['consume'] = 'enrich'
        exp_phase_route['enrich'] = 'publish'
        exp_phase_map['enrich'] = interfaces.IMessageHandler
    else:
        inited_plugins.pop('enricher')

    router_inst = router.GordonRouter(
        exp_phase_route, plugins=inited_plugins.values(), metrics=metrics,
        **plugin_kwargs)

    assert exp_phase_route == router_inst.phase_route
    for phase, plugin in router_inst.phase_plugin_map.items():
        assert exp_phase_map[phase].providedBy(plugin)


@pytest.mark.parametrize('current_phase,exp_next_phase', (
    ('consume', 'enrich'),
    ('its_not_a_phase_mom', 'cleanup'),
))
def test_get_next_phase(current_phase, exp_next_phase, router_inst, event_msg,
                        caplog):
    event_msg.phase = current_phase
    next_phase = router_inst._get_next_phase(event_msg)

    assert exp_next_phase == next_phase
    assert 1 == len(caplog.records)


@pytest.mark.asyncio
@pytest.mark.parametrize('msg_exists,exp_start_count', (
    (True, 0),
    (False, 1),
))
async def test_add_message_in_flight(msg_exists, exp_start_count, router_inst,
                                     event_msg):
    if msg_exists:
        timer = router_inst.metrics.timer(
            'router-message-flight-duration', {'unit': 'seconds'})
        router_inst._messages_in_flight[event_msg.msg_id] = timer

    await router_inst._add_message_in_flight(event_msg)

    assert event_msg.msg_id in router_inst._messages_in_flight

    if not msg_exists:
        timer = router_inst._messages_in_flight[event_msg.msg_id]

    assert exp_start_count == timer._mock_start_call_count

    assert 'router-message-flight-duration' == timer.metric_name
    assert 1 == router_inst.metrics._mock_set_call_count

    exp_args = [('router-messages-in-flight', 1, None, {})]
    assert exp_args == router_inst.metrics._mock_set_call_args


@pytest.mark.asyncio
@pytest.mark.parametrize('phase,exp_stop_count,exp_msg_count', (
    ('cleanup', 1, 0),
    ('consume', 0, 1),
))
async def test_remove_message_in_flight(phase, exp_stop_count, exp_msg_count,
                                        router_inst, event_msg):
    event_msg.phase = phase
    timer = router_inst.metrics.timer(
        'router-message-flight-duration', {'unit': 'seconds'})
    router_inst._messages_in_flight[event_msg.msg_id] = timer

    await router_inst._remove_message_in_flight(event_msg)

    assert exp_msg_count == len(router_inst._messages_in_flight)
    assert exp_stop_count == timer._mock_stop_call_count
    assert 'router-message-flight-duration' == timer.metric_name
    assert 1 == router_inst.metrics._mock_set_call_count

    exp_args = [('router-messages-in-flight', exp_msg_count, None, {})]
    assert exp_args == router_inst.metrics._mock_set_call_args


@pytest.mark.asyncio
@pytest.mark.parametrize('raises,exp_call_count,exp_log_count', (
    (False, 0, 1),
    (True, 1, 2),
))
async def test_route(raises, exp_call_count, exp_log_count, event_msg,
                     router_inst, caplog, monkeypatch):

    mock_process_call_count = 0
    mock_process_call_args = []

    async def mock_process(*args, **kwargs):
        nonlocal mock_process_call_count
        nonlocal mock_process_call_args

        mock_process_call_count += 1
        mock_process_call_args.append((args, kwargs))
        if raises:
            raise Exception('foo')

    monkeypatch.setattr(
        router_inst.phase_plugin_map['enrich'], 'handle_message', mock_process)

    await router_inst._route(event_msg)

    assert 1 == mock_process_call_count
    assert [((event_msg,), {})] == mock_process_call_args
    assert 1 == event_msg._mock_append_to_history
    assert exp_log_count == len(caplog.records)
    actual_count = \
        router_inst.phase_plugin_map['cleanup']._mock_handle_message_count
    assert exp_call_count == actual_count


@pytest.mark.asyncio
async def test_route_no_phase_no_cleanup(event_msg, router_inst, caplog):
    del(router_inst.phase_plugin_map['cleanup'])
    event_msg.phase = 'blahblah'  # doesnt exist

    await router_inst._route(event_msg)
    assert 2 == len(caplog.records)  # unknown phase, drop


@pytest.mark.asyncio
@pytest.mark.parametrize('phase,raises,incr_count,incr_args', (
    ('consume', False, 1, [('router-update-message-phase', 1,
                            {'current': 'consume', 'next': 'enrich'},
                            {}), ]),
    ('publish', False, 2, [('router-update-message-phase', 1,
                            {'current': 'publish', 'next': 'cleanup'}, {}),
                           ('router-message-completed', 1, None, {})]),
    ('consume', True, 3, [('router-update-message-phase', 1,
                           {'current': 'consume', 'next': 'enrich'}, {}),
                          ('router-message-dropped', 1,
                           {'error': 'Exception'}, {}),
                          ('router-update-message-phase', 1,
                           {'current': 'enrich', 'next': 'cleanup'}, {})]),
))
async def test_route_metrics(phase, raises, incr_count, incr_args, router_inst,
                             event_msg, monkeypatch):
    """Various metrics are called when message is routed."""
    event_msg.phase = phase

    async def mock_handle_message(*args, **kwargs):
        if raises:
            raise Exception('foo')

    monkeypatch.setattr(
        router_inst.phase_plugin_map['enrich'], 'handle_message',
        mock_handle_message)

    await router_inst._route(event_msg)

    assert incr_count == router_inst.metrics._mock_incr_call_count
    assert incr_args == router_inst.metrics._mock_incr_call_args


@pytest.mark.asyncio
@pytest.mark.parametrize('enricher,raises', (
    (True, False),
    (True, True),
    (False, True),
    (False, False),
))
async def test_poll_channel(enricher, raises, router_inst, event_msg, caplog,
                            monkeypatch):
    if not enricher:
        router_inst.phase_route.pop('enrich')
        router_inst.phase_route['consume'] = 'publish'

    if raises:
        event_msg.phase = 'its_not_a_phase_mom'

    mock_route_call_count = 0
    mock_route_call_args = []

    async def mock_route(*args, **kwargs):
        nonlocal mock_route_call_count
        nonlocal mock_route_call_args

        mock_route_call_count += 1
        mock_route_call_args.append((args, kwargs))

    monkeypatch.setattr(router_inst, '_route', mock_route)

    await router_inst.success_channel.put(event_msg)
    await router_inst.success_channel.put(None)
    await router_inst._poll_channel()

    assert 1 == mock_route_call_count
    assert [((event_msg,), {})] == mock_route_call_args


@pytest.mark.asyncio
async def test_poll_channel_metrics(router_inst, event_msg, monkeypatch):
    """Various metrics are called when message is polled."""

    async def mock_route(*args, **kwargs):
        pass

    monkeypatch.setattr(router_inst, '_route', mock_route)

    mock_add_call_count = 0
    mock_add_call_args = []

    async def mock_add_message_in_flight(*args, **kwargs):
        nonlocal mock_add_call_count
        nonlocal mock_add_call_args

        mock_add_call_count += 1
        mock_add_call_args.append((args, kwargs))

    monkeypatch.setattr(
        router_inst, '_add_message_in_flight',
        mock_add_message_in_flight)

    mock_remove_call_count = 0
    mock_remove_call_args = []

    async def mock_remove_message_in_flight(*args, **kwargs):
        nonlocal mock_remove_call_count
        nonlocal mock_remove_call_args

        mock_remove_call_count += 1
        mock_remove_call_args.append((args, kwargs))

    monkeypatch.setattr(
        router_inst, '_remove_message_in_flight',
        mock_remove_message_in_flight)

    await router_inst.success_channel.put(event_msg)
    await router_inst.success_channel.put(None)

    await router_inst._poll_channel()

    assert 1 == mock_add_call_count
    assert 1 == mock_remove_call_count

    args = [((event_msg,), {}), ]
    assert args == mock_add_call_args
    assert args == mock_remove_call_args

    assert 1 == router_inst.metrics._mock_incr_call_count
    exp_args = [('router-message-consumed', 1, None, {})]
    assert exp_args == router_inst.metrics._mock_incr_call_args


@pytest.mark.asyncio
async def test_poll_channel_invalid_msg(router_inst, caplog, mocker,
                                        monkeypatch):
    event_msg = mocker.Mock()
    mock_next_phase = mocker.Mock()
    monkeypatch.setattr(router_inst, '_get_next_phase', mock_next_phase)
    await router_inst.success_channel.put(event_msg)

    await router_inst._poll_channel()

    assert 1 == len(caplog.records)
    mock_next_phase.assert_not_called()
    assert 1 == router_inst.metrics._mock_incr_call_count

    exp = (
        'router-message-dropped', 1, {'error': 'invalid-message-provider'}, {}
    )
    assert [exp, ] == router_inst.metrics._mock_incr_call_args


@pytest.fixture
def event_loop_mock(mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr('gordon.router.asyncio.get_event_loop', mock)
    return mock.return_value


@pytest.mark.asyncio
async def test_run(caplog, event_loop_mock, router_inst, mocker, monkeypatch):
    poll_mock = mocker.Mock()
    monkeypatch.setattr(router_inst, '_poll_channel', poll_mock)

    await router_inst.run()

    event_loop_mock.create_task.assert_called_once_with(poll_mock.return_value)
    assert 1 == len(caplog.records)



# -*- coding: utf-8 -*-
# Copyright (c) 2018 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

Gordon ships with a simple `ffwd <https://github.com/spotify/ffwd>`_
metrics implementation, which can be enabled via configuration. This
module contains the SimpleFfwdRelay, and all required classes that it
uses to send messsages to the ffwd daemon via UDP.

The `SimpleFfwdRelay` requires no configuration, but can be customized.
The defaults that may be overridden are shown below.

.. code-block:: ini

    [ffwd]
    # to identify the service creating metrics
    key = 'gordon-service'

    # the address of the ffwd daemon (see: UDPClient)
    ip = "127.0.0.9"
    port = 19000

    # a scaling factor for timing (see: FfwdTimer)
    time_unit = 1E9

"""

import asyncio
import json
import time

import zope.interface

from gordon import interfaces


class UDPClientProtocol(asyncio.DatagramProtocol):
    """Protocol for sending one-off messages via UDP.

    Args:
        message (bytes): Message for ffwd agent.
    """
    def __init__(self, message):
        self.message = message
        self.transport = None

    def connection_made(self, transport):
        """Create connection, use to send message and close.

        Args:
            transport (asyncio.DatagramTransport): Transport used for sending.
        """
        self.transport = transport
        self.transport.sendto(self.message)
        self.transport.close()


class UDPClient:
    """Client for sending UDP datagrams.

    Args:
        ip (str): (optional) Destination IP address (default: 127.0.0.1).
        port (int): (optional) Destination port (default: 9000).
        loop (asyncio.AbstractEventLoop impl): (optional) Event loop.
    """

    DEFAULT_IP = '127.0.0.1'
    DEFAULT_PORT = 19000

    def __init__(self, ip=None, port=None, loop=None):
        self.ip = ip or self.DEFAULT_IP
        self.port = port or self.DEFAULT_PORT
        self.loop = loop or asyncio.get_event_loop()

    async def send(self, metric):
        """Transform metric to JSON bytestring and send to server.

        Args:
            metric (dict): Complete metric to send as JSON.
        """
        message = json.dumps(metric).encode('utf-8')
        await self.loop.create_datagram_endpoint(
            lambda: UDPClientProtocol(message),
            remote_addr=(self.ip, self.port))


@zope.interface.implementer(interfaces.ITimer)
class FfwdTimer:
    """Timer which sends UDP messages to FFWD on completion.

    Args:
        metric (dict): Dict representation of the metric to send.
        udp_client (UDPClient): A metric sending client.
        time_unit (number): (optional) Scale time unit for use with
            time.perf_counter(), for example: 1E9 to send nanoseconds.
    """
    def __init__(self, metric, udp_client, time_unit=None):
        self.metric = metric
        self.udp_client = udp_client
        self.time_unit = time_unit or 1

    async def __aenter__(self):
        """Enter context manager to start timing."""
        await self.start()
        return self

    async def __aexit__(self, *args):
        """Exit context manager to stop timing."""
        await self.stop()

    async def start(self):
        """Start timer."""
        self._start_time = time.perf_counter()

    async def stop(self):
        """Stop timer."""
        time_elapsed = time.perf_counter() - self._start_time
        self.metric['value'] = time_elapsed * self.time_unit
        await self.udp_client.send(self.metric)


@zope.interface.implementer(interfaces.IMetricRelay)
class SimpleFfwdRelay:
    """Metrics relay which sends to FFWD immediately.

    The relay does no client-side aggregation and metrics are
    emitted immediately. The relay uses a combination of the `key and
    attributes fields <https://github.com/spotify/ffwd/tree/master/
    modules/json>`_ to semantically identify metrics in ffwd.

    Args:
        config (dict): Configuration with optional keys described above.
    """
    def __init__(self, config):
        self.key = config.get('key', 'gordon-service')
        self.udp_client = UDPClient(
            config.get('ffwd_ip'), config.get('ffwd_port'))
        self.time_unit = config.get('time_unit')

    def _create_metric(self, metric_name, value, context, **kwargs):
        attrs = context or {}
        attrs['what'] = metric_name

        metric = {
            'key': self.key,
            'attributes': attrs,
            'value': value,
            'type': 'metric'
        }
        return metric

    async def incr(self, metric_name, value=1, context=None, **kwargs):
        """Increase a metric by 1 or a given amount.

        Args:
            metric_name (str): Identifier of the metric.
            value (int): (optional) Value with which to increase the metric
                (default: 1).
            context (dict): (optional) Additional key-value pairs which further
                describe the metric, for example: {'remote-host': '1.2.3.4'}
        """
        await self.set(metric_name, value, context, **kwargs)

    def timer(self, metric_name, context=None, **kwargs):
        """Create a FfwdTimer.

        Args:
            metric_name (str): Identifier of the metric.
            context (dict): (optional) Additional key-value pairs which further
                describe the metric, for example: {'unit': 'seconds'}
        """
        metric = self._create_metric(metric_name, None, context, **kwargs)
        return FfwdTimer(metric, self.udp_client, self.time_unit)

    async def set(self, metric_name, value, context=None, **kwargs):
        """Set a metric to a given value.

        Args:
            metric_name (str): Identifier of the metric.
            value (number): The value of the metric.
            context (dict): (optional) Additional key-value pairs which further
                describe the metric, for example: {'app-version': '1.5.3'}
        """
        metric = self._create_metric(metric_name, value, context, **kwargs)
        await self.udp_client.send(metric)

    async def cleanup(self):
        """Not used."""
        pass



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import logging
import time

import async_dns
from async_dns.resolver import ProxyResolver

from gordon import exceptions


class RecordChecker(object):

    def __init__(self, dns_ip):
        """Setup RecordChecker object.

        Args:
            dns_ip: DNS server IP to query.
        """
        self._dns_ip = dns_ip
        self._resolver = ProxyResolver()
        try:
            self._resolver.set_proxies([self._dns_ip])
        except async_dns.address.InvalidHost as e:
            msg = f'RecordChecker got invalid DNS server IP: {e}.'
            raise exceptions.InvalidDNSHost(msg)

    def _extract_record_data(self, record):
        """Extract record data from a record dict.

        Args:
            record (dict): DNS record as a dict.

        Returns:
            The record data: name, rrdata, type and ttl.
        """
        return record.values()

    async def check_record(self, record, timeout=60):
        """Measures the time for a DNS record to become available.

        Query a provided DNS server multiple times until the reply matches the
        information in the record or until timeout is reached.

        Args:
            record (dict): DNS record as a dict with record properties.
            timeout (int): Time threshold to query the DNS server.
        """
        start_time = time.time()

        name, rr_data, r_type, ttl = self._extract_record_data(record)
        r_type_code = async_dns.types.get_code(r_type)

        resolvable_record = False
        retries = 0
        sleep_time = 5

        while not resolvable_record and \
                timeout > retries * sleep_time:

            retries += 1
            resolver_res = await self._resolver.query(name, r_type_code)
            possible_ans = resolver_res.an

            resolvable_record = \
                await self._check_resolver_ans(possible_ans, name,
                                               rr_data, ttl, r_type_code)

            if not resolvable_record:
                await asyncio.sleep(sleep_time)

        if not resolvable_record:
            logging.info(
                f'Sending metric record-checker-failed: {record}.')
        else:
            final_time = float(time.time() - start_time)
            success_msg = (f'This record: {record} took {final_time} to '
                           'register.')
            logging.info(success_msg)

    async def _check_resolver_ans(
            self, dns_answer_list, record_name,
            record_data_list, record_ttl, record_type_code):
        """Check if resolver answer is equal to record data.

        Args:
            dns_answer_list (list): DNS answer list contains record objects.
            record_name (str): Record name.
            record_data_list (list): List of data values for the record.
            record_ttl (int): Record time-to-live info.
            record_type_code (int): Record type code.

        Returns:
            boolean indicating if DNS answer data is equal to record data.
        """
        type_filtered_list = [
            ans for ans in dns_answer_list if ans.qtype == record_type_code
        ]

        # check to see that type_filtered_lst has
        # the same number of records as record_data_list
        if len(type_filtered_list) != len(record_data_list):
            return False

        # check each record data is equal to the given data
        for rec in type_filtered_list:
            conditions = [rec.name == record_name,
                          rec.ttl == record_ttl,
                          rec.data in record_data_list]

            # if ans record data is not equal
            # to the given data return False
            if not all(conditions):
                return False

        return True



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from async_dns import Record

from gordon import exceptions, record_checker


#####
# Tests & fixtures for running record checker
#####

@pytest.fixture
def record_checker_instance():
    dns_ip = '8.8.8.8'
    return record_checker.RecordChecker(dns_ip)


@pytest.fixture
def record_1(mocker):
    record_mock = mocker.Mock(Record)
    record_mock.qtype = 1
    record_mock.name = 'example.com'
    record_mock.data = '127.1.1.1'
    record_mock.ttl = 10686
    return record_mock


@pytest.fixture
def record_2(record_1):
    record_1.data = '127.1.1.2'
    return record_1


@pytest.fixture
def record_3(record_1):
    record_1.data = '127.1.1.3'
    return record_1


@pytest.fixture
def dns_query_response(mocker, record_1, record_2, record_3):
    dns_res_mock = mocker.Mock()
    dns_res_mock.an = [record_1, record_2, record_3]
    return dns_res_mock


@pytest.fixture
def get_mock_coro(mocker):
    """Create a mock-coro pair.
    The coro can be used to patch an async method while the mock can
    be used to assert calls to the mocked out method.
    """
    def _create_mock_coro_pair():
        mock = mocker.Mock()

        async def _coro(*args, **kwargs):
            return mock(*args, **kwargs)
        return mock, _coro
    return _create_mock_coro_pair


@pytest.fixture
def mock_resolve_query(mocker, get_mock_coro, record_checker_instance):
    mock, _coroutine = get_mock_coro()
    mocker.patch.object(record_checker_instance._resolver, 'query', _coroutine)
    return mock


@pytest.mark.asyncio
async def test_check_time_on_success(
        record_checker_instance, caplog,
        mock_resolve_query, dns_query_response):
    """Test the time it took to check a record is in expected range"""
    record_to_check = {
        'name': 'example.com',
        'rrdatas': ['127.1.1.1', '127.1.1.2', '127.1.1.3'],
        'type': 'A',
        'ttl': 10686
    }

    mock_resolve_query.return_value = dns_query_response

    await record_checker_instance.check_record(record_to_check)

    success_msg = caplog.records[0].msg

    time_it_took_to_check_record = \
        float(success_msg.split('took ')[1].split(' to')[0])

    max_time_to_sleep_when_success = 6

    assert 0 < time_it_took_to_check_record < max_time_to_sleep_when_success


@pytest.mark.asyncio
async def test_check_record_failure_ttl_not_equal(
        mocker, get_mock_coro,
        record_checker_instance, caplog,
        mock_resolve_query, dns_query_response):
    """Test failure timeout.

    The failure happens because the ttl is not equal.
    """
    record_to_check = {
        'name': 'example.com',
        'rrdatas': ['127.1.1.1', '127.1.1.2', '127.1.1.3'],
        'type': 'A',
        'ttl': 1068
    }

    mock_resolve_query.return_value = dns_query_response

    mock, _coroutine = get_mock_coro()
    mocker.patch('asyncio.sleep', _coroutine)

    await record_checker_instance.check_record(record_to_check)
    actual_msg = caplog.records[0].msg

    expected_msg = f'Sending metric record-checker-failed: {record_to_check}.'
    assert expected_msg == actual_msg


@pytest.mark.asyncio
async def test_length_of_rrdata_is_not_equal(
        mocker, get_mock_coro,
        record_checker_instance, caplog,
        mock_resolve_query, dns_query_response):
    """Test failure timeout.

    The failure happens because the length of
        rrdatas list is not equal.
    """
    record_to_check = {
        'name': 'example.com',
        'rrdatas': ['127.1.1.1', '127.1.1.2'],
        'type': 'A',
        'ttl': 10686
    }

    mock_resolve_query.return_value = dns_query_response

    mock, _coroutine = get_mock_coro()
    mocker.patch('asyncio.sleep', _coroutine)

    await record_checker_instance.check_record(record_to_check)
    actual_msg = caplog.records[0].msg

    expected_msg = f'Sending metric record-checker-failed: {record_to_check}.'
    assert expected_msg == actual_msg


@pytest.mark.asyncio
async def test_proxy_resolver_failure():
    """Test record checker got invalid dns ip"""
    dns_ip = '8.855.0.5'

    with pytest.raises(exceptions.InvalidDNSHost) as e:
        record_checker.RecordChecker(dns_ip)

    expected_msg = 'RecordChecker got invalid DNS server IP: 8.855.0.5.'
    assert e.match(expected_msg)



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import signal

import pytest
from click.testing import CliRunner

from gordon import main
from tests.unit import conftest


#####
# Tests for service shutdown
#####
@pytest.mark.asyncio
async def test_shutdown(mocker, monkeypatch, caplog, event_loop):
    async def foo():
        await asyncio.sleep(0)

    task = asyncio.Task(foo())
    mock_task = mocker.Mock()
    mock_task.all_tasks.return_value = [task]
    monkeypatch.setattr('gordon.main.asyncio.Task', mock_task)

    await main.shutdown(signal.SIGTERM, event_loop)

    assert 4 == len(caplog.records)
    assert task.cancelled()


#####
# Tests for service setup
#####
@pytest.mark.parametrize('suffix', ['', '-user'])
def test_load_config(tmpdir, suffix, config_file, loaded_config):
    """Load prod and user config."""
    conf_file = tmpdir.mkdir('config').join('gordon{}.toml'.format(suffix))
    conf_file.write(config_file)
    config = main._load_config(root=conf_file.dirpath())
    assert loaded_config == config


def test_load_config_raises(tmpdir):
    """No config loaded raises IOError."""
    dir_with_no_conf = tmpdir.mkdir('config')
    with pytest.raises(IOError) as e:
        main._load_config(root=dir_with_no_conf.dirpath())

    assert e.match('Cannot load Gordon configuration file from')


def test_setup(tmpdir, mocker, monkeypatch, config_file, loaded_config):
    """Setup service config and logging."""
    conf_file = tmpdir.mkdir('config').join('gordon.toml')
    conf_file.write(config_file)

    ulogger_mock = mocker.MagicMock(main.ulogger, autospec=True)
    ulogger_mock.setup_logging = mocker.Mock()
    monkeypatch.setattr(main, 'ulogger', ulogger_mock)

    config = main.setup(config_root=conf_file.dirpath())

    assert loaded_config == config

    ulogger_mock.setup_logging.assert_called_once_with(
        progname='gordon', level='DEBUG', handlers=['stream'])


#####
# Tests & fixtures for running service
#####
@pytest.fixture
def setup_mock(mocker, monkeypatch):
    setup_mock = mocker.MagicMock(main.setup, autospec=True)
    monkeypatch.setattr(main, 'setup', setup_mock)
    return setup_mock


@pytest.fixture
def mock_plugins_loader(mocker, monkeypatch):
    mock_plugins_loader = mocker.MagicMock(
        main.plugins_loader.load_plugins, autospec=True)
    monkeypatch.setattr(
        main.plugins_loader, 'load_plugins', mock_plugins_loader)
    return mock_plugins_loader


def test_log_or_exit_on_exceptions_no_debug(plugin_exc_mock, mocker,
                                            monkeypatch):
    """Raise SystemExit if debug flag is off."""
    logging_mock = mocker.MagicMock(main.logging, autospec=True)
    monkeypatch.setattr(main, 'logging', logging_mock)

    errors = [('bad.plugin', plugin_exc_mock)]
    with pytest.raises(SystemExit) as e:
        main._log_or_exit_on_exceptions('base msg', errors, debug=False)

    e.match('1')
    logging_mock.error.assert_called_once()
    logging_mock.warn.assert_not_called()


def test_log_or_exit_on_exceptions_debug(plugin_exc_mock, mocker, monkeypatch):
    """Do not exit out if debug flag is on."""
    logging_mock = mocker.MagicMock(main.logging, autospec=True)
    monkeypatch.setattr(main, 'logging', logging_mock)

    errors = [('bad.plugin', plugin_exc_mock)]

    main._log_or_exit_on_exceptions('base msg', errors, debug=True)

    logging_mock.warn.assert_called_once()
    logging_mock.error.assert_not_called()


@pytest.fixture
def mock_log_or_exit_on_exc(mocker, monkeypatch):
    mock = mocker.Mock()
    monkeypatch.setattr('gordon.main._log_or_exit_on_exceptions', mock)
    return mock


@pytest.mark.parametrize('plugins,exp_log_calls', (
    (('event_consumer', 'enricher', 'publisher', 'runnable'), 0),
    (('event_consumer', 'enricher', 'publisher'), 0),
    (('runnable', 'enricher'), 0),
    (('runnable',), 1),
    (('event_consumer',), 0),
    (('enricher',), 1),

))
def test_assert_required_plugins(mocker, plugins, exp_log_calls,
                                 inited_plugins, mock_log_or_exit_on_exc):
    """Assert required plugins are installed, else warn/error out."""
    plugins = [inited_plugins[p] for p in plugins]
    main._gather_plugins_by_type(plugins, debug=True)

    assert exp_log_calls == mock_log_or_exit_on_exc.call_count


@pytest.mark.parametrize('patches,exp_mock_call', (
    # provider: async run; gen plugin: async run
    ([], 0),
    # provider: sync run; gen plugin: async run
    ([('EventConsumerStub.run', 'set')], 1),
    # provider: no run; gen plugin: async run
    ([('EventConsumerStub.run', 'del')], 1),
    # provider: sync run; gen plugin: sync run
    ([('EventConsumerStub.run', 'set'), ('GenericStub.run', 'set')], 3),
    # provider: no run; gen plugin: sync run
    ([('EventConsumerStub.run', 'del'), ('GenericStub.run', 'set')], 3),
    # provider: sync run; gen plugin: no run
    ([('EventConsumerStub.run', 'set'), ('GenericStub.run', 'del')], 3),
    # provider: no run; gen plugin: no run
    ([('EventConsumerStub.run', 'del'), ('GenericStub.run', 'del')], 3),
    # provider: async run; gen plugin: sync run
    ([('GenericStub.run', 'set')], 1),
    # provider: async run; gen plugin: no run
    ([('GenericStub.run', 'del')], 1),
    # confirm the same logic works for handle_message
    ([('EventConsumerStub.handle_message', 'set')], 1),
    ([('EventConsumerStub.handle_message', 'del')], 1),
))
def test_gather_plugins_by_type_logs(patches, exp_mock_call, monkeypatch,
                                     inited_plugins, mock_log_or_exit_on_exc):

    def _set_or_delete(patch, action):
        if action == 'del':
            monkeypatch.delattr(patch, raising=False)
        else:
            monkeypatch.setattr(patch, lambda: None)

    if patches:
        base_patch = 'tests.unit.conftest.'
        for patch, action in patches:
            patch = base_patch + patch
            _set_or_delete(patch, action)

    main._gather_plugins_by_type(inited_plugins.values(), debug=True)

    assert exp_mock_call == mock_log_or_exit_on_exc.call_count


@pytest.mark.parametrize('input_plugins,exp_mock_call,exp_plugin_count', [
    (['event_consumer'], 0, (1, 1)),
    (['event_consumer', 'publisher'], 0, (1, 2)),
    (['publisher'], 1, (0, 1)),
    (['runnable'], 1, (1, 0)),
])
def test_gather_plugins_by_type(input_plugins, exp_mock_call, exp_plugin_count,
                                inited_plugins, mock_log_or_exit_on_exc):
    plugins = [inited_plugins[i] for i in input_plugins]
    runnable_plugins, message_handlers = \
        main._gather_plugins_by_type(plugins, debug=True)
    assert exp_plugin_count == (len(runnable_plugins), len(message_handlers))
    assert exp_mock_call == mock_log_or_exit_on_exc.call_count


@pytest.mark.asyncio
async def test_run_plugins(inited_plugins, mocker, monkeypatch):
    """Run all installed plugins."""
    async def mock_run(*args, **kwargs):
        await asyncio.sleep(0)

    mock_router = mocker.Mock()
    monkeypatch.setattr(mock_router, 'run', mock_run)

    runnable_plugins, _ = main._gather_plugins_by_type(
        inited_plugins.values(), True)
    await main._run(runnable_plugins, mock_router, debug=True)

    assert 1 == inited_plugins['event_consumer']._mock_run_count
    assert 1 == inited_plugins['runnable']._mock_run_count


@pytest.fixture
def event_loop_mock(mocker, monkeypatch):
    mock = mocker.Mock()
    mock.return_value.add_signal_handler = mocker.Mock()
    monkeypatch.setattr('gordon.main.asyncio.get_event_loop', mock)
    return mock.return_value


@pytest.mark.parametrize('has_active_plugins,exp_log_count,errors', (
    (True, 2, []),
    (True, 2, [('not_a.plugin', conftest.plugin_exc_mock())]),
    (False, 2, []),
))
def test_run_cli(has_active_plugins, exp_log_count, errors, installed_plugins,
                 setup_mock, mock_plugins_loader, mocker, monkeypatch, caplog,
                 loaded_config, event_loop_mock):
    """Successfully run gordon service in debug mode via CLI."""
    loaded_config['core']['debug'] = True
    names, _plugins = [], []

    if has_active_plugins:
        names = [
            'event_consumer.plugin',
            'enricher.plugin',
            'publisher.plugin'
        ]
        _plugins = [
            conftest.EventConsumerStub({}, mocker.Mock(), mocker.Mock()),
            conftest.EnricherStub({}, None),
            conftest.PublisherStub({}, None)
        ]
    mock_plugins_loader.return_value = names, _plugins, errors, mocker.Mock()
    _run_mock = mocker.Mock()
    monkeypatch.setattr('gordon.main._run', _run_mock)
    _log_or_exit_mock = mocker.Mock()
    monkeypatch.setattr('gordon.main._log_or_exit_on_exceptions',
                        _log_or_exit_mock)

    runner = CliRunner()
    result = runner.invoke(main.run)

    assert 0 == result.exit_code
    setup_mock.assert_called_once()
    assert 3 == event_loop_mock.add_signal_handler.call_count
    event_loop_mock.create_task.assert_called_once_with(
        _run_mock(_plugins, True))
    event_loop_mock.run_forever.assert_called_once_with()
    event_loop_mock.close.assert_called_once()
    assert exp_log_count == len(caplog.records)
    if errors:
        _log_or_exit_mock.assert_called_once()


def test_run_cli_raise_exceptions(loaded_config, installed_plugins, caplog,
                                  setup_mock, mock_plugins_loader,
                                  plugin_exc_mock, mocker):
    """Raise plugin exceptions when not in debug mode via CLI."""
    loaded_config['core']['debug'] = False
    setup_mock.return_value = loaded_config

    names = ['event_consumer', 'enricher']
    errors = [('not_a_plugin', plugin_exc_mock)]
    mock_plugins_loader.return_value = names, installed_plugins, errors, \
        mocker.Mock()

    runner = CliRunner()
    result = runner.invoke(main.run)

    assert 1 == result.exit_code
    setup_mock.assert_called_once()
    assert 1 == len(caplog.records)



# -*- coding: utf-8 -*-
#
# Copyright 2018 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Core message routing logic for the plugins within Gordon Service.

Messages received on the success channel will be routed to the next
designated plugin phase. For example, a message that has a ``consume``
phase will be routed to the installed enricher provider (or publisher
provider if no enricher provider is installed).

If a message fails its next phase, its phase will be updated to ``drop``
and routed to the event consumer provider for cleanup.

.. attention::

    The :class:`GordonRouter` only supports the following two phase
    routes:

    1. consume -> enrich -> publish -> done
    2. consume -> publish -> done

    Future releases may support more configurable phase routings.

"""

import asyncio
import logging

from gordon import interfaces


class GordonRouter:
    """Route messages to the appropriate plugin destination.

    .. attention::

        `error_channel` is currently not used in this router, and may
        be removed entirely from all interface definitions.

    Args:
        phase_route (dict(str, str)): The route messages should follow.
        success_channel (asyncio.Queue): A sink for successfully
            processed :class:`gordon.interfaces.IEventMessage` s.
        error_channel (asyncio.Queue): A sink for
            :class:`gordon.interfaces.IEventMessage` s that were not
            processed due to problems.
        plugins (list): Instantiated message handling plugins.
        metrics (obj): Implemented :class:`IMetricRelay` interface
            to emit metrics.
    """
    # TODO (lynn): Ideally this is configurable/extendable in future
    #              iterations of gordon.
    FINAL_PHASES = ('cleanup',)

    def __init__(self, phase_route, success_channel, error_channel, plugins,
                 metrics):
        self.success_channel = success_channel
        self.error_channel = error_channel
        self.phase_plugin_map = self._get_phase_plugin_map(plugins)
        self.phase_route = phase_route
        self.metrics = metrics
        self._messages_in_flight = {}

    def _get_phase_plugin_map(self, plugins):
        phase_map = {p.phase: p for p in plugins}
        if not set(self.FINAL_PHASES) & set(phase_map):
            msg = (f'None of {self.FINAL_PHASES} implemented, will default'
                   'to dropping messages at these phases.')
            logging.warn(msg)
        return phase_map

    def _get_next_phase(self, event_msg):
        try:
            next_phase = self.phase_route[event_msg.phase]
            msg = f'Routing message {event_msg.msg_id} to "{next_phase}".'
            logging.debug(msg)

        except KeyError:
            msg = (f'Message "{event_msg.msg_id}" has an unknown phase: '
                   f'"{event_msg.phase}", routing to "cleanup".')
            logging.error(msg)
            next_phase = 'cleanup'
        return next_phase

    async def _update_messages_in_flight(self):
        in_flight_count = len(self._messages_in_flight)
        await self.metrics.set(
            'router-messages-in-flight', value=in_flight_count)

    async def _add_message_in_flight(self, event_msg):
        if event_msg.msg_id not in self._messages_in_flight:
            context = {'unit': 'seconds'}
            timer = self.metrics.timer(
                'router-message-flight-duration', context=context)
            await timer.start()
            self._messages_in_flight[event_msg.msg_id] = timer

        await self._update_messages_in_flight()

    async def _remove_message_in_flight(self, event_msg):
        if event_msg.phase in self.FINAL_PHASES:
            timer = self._messages_in_flight.pop(event_msg.msg_id)
            await timer.stop()

        await self._update_messages_in_flight()

    async def _route(self, event_msg, force_cleanup=False):
        if force_cleanup:
            next_phase = 'cleanup'
        else:
            next_phase = self._get_next_phase(event_msg)

        context = {
            'current': event_msg.phase,
            'next': next_phase,
        }
        await self.metrics.incr('router-update-message-phase', context=context)

        event_msg.update_phase(next_phase)
        next_plugin = self.phase_plugin_map.get(next_phase)
        if next_phase in self.FINAL_PHASES and not next_plugin:
            msg = (f'Dropping message "{event_msg.msg_id}", final phase '
                   f'"{next_phase}" not implemented.')
            logging.debug(msg)
            return

        try:
            await next_plugin.handle_message(event_msg)
            # don't add a successfully dropped message back onto channel
            if next_phase not in self.FINAL_PHASES:
                msg = f'Adding message "{event_msg}" to success channel.'
                event_msg.append_to_history(msg, event_msg.phase)
                await self.success_channel.put(event_msg)
            elif next_phase in self.FINAL_PHASES and not force_cleanup:
                # if we're here, can assume message successfully went thru
                # the entire phase route and not dropped
                await self.metrics.incr('router-message-completed')

        except Exception as e:
            msg = (f'Routing message "{event_msg}" to cleanup due to exception:'
                   f' "{e}"')
            event_msg.append_to_history(msg, event_msg.phase)
            logging.warn(msg, exc_info=e)
            context = {'error': e.__class__.__name__}
            await self.metrics.incr('router-message-dropped', context=context)
            await self._route(event_msg, force_cleanup=True)

    async def _verify_message_impl(self, event_msg):
        if not interfaces.IEventMessage.providedBy(event_msg):
            msg = (f'Ignoring message "{event_msg.msg_id}". Does not correctly'
                   ' implement `IEventMessage`.')
            logging.warn(msg)

            context = {'error': 'invalid-message-provider'}
            await self.metrics.incr('router-message-dropped', context=context)
            return False
        return True

    async def _poll_channel(self):
        while True:
            event_msg = await self.success_channel.get()

            # graceful shutdown
            if event_msg is None:
                msg = 'Received TERM signal, shutting down router...'
                logging.info(msg)
                # TODO (lynn): potentially propagate signal all plugins
                #              to clean up rather than just breaking
                break

            if not await self._verify_message_impl(event_msg):
                break

            await self.metrics.incr('router-message-consumed')
            await self._add_message_in_flight(event_msg)
            await self._route(event_msg)
            await self._remove_message_in_flight(event_msg)

    async def run(self):
        """Entrypoint to route messages between plugins."""
        logging.info('Starting message router...')
        loop = asyncio.get_event_loop()
        loop.create_task(self._poll_channel())



# -*- coding: utf-8 -*-
# Copyright (c) 2018 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Gordon ships with a default IMetricRelay implementation, which
outputs metrics to the application logger using the standard
library `logging` module.

This implementation is used by default if no other IMetricRelay
implementation is loaded and chosen.

The default configuration that can be overriden is shown below.

.. code-block:: ini

    [metrics-logger]
    # One of the log levels available from the stdlib logging module.
    log_level = 'info'
    # A scaling factor for timing (see: LogTimer)
    time_unit = 1
"""

import collections
import logging
import time

import zope.interface

from gordon import interfaces


@zope.interface.implementer(interfaces.ITimer)
class LogTimer:
    """Timer that outputs metrics to the logger.

    Args:
        metric (dict): Metric to output.
        logger (.LoggerAdapter): Object used for outputting metrics.
        time_unit (number): (optional) Scale time unit for use with
            time.perf_counter(), for example: 1E9 to send nanoseconds.
    """
    def __init__(self, metric, logger, time_unit=1):
        self.metric = metric
        self.logger = logger
        self.time_unit = time_unit

    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, *args):
        await self.stop()

    async def start(self):
        """Start the timer."""
        self._start_time = time.perf_counter()

    async def stop(self):
        """Stop the timer and output the time delta to the logger."""
        time_elapsed = (
            time.perf_counter() - self._start_time) * self.time_unit
        self.metric['value'] = time_elapsed
        self.logger.log(self.metric)


class LoggerAdapter:
    """Configurable adapter that formats and outputs metrics.

    Args:
        level (str): Log level at which to output metric.
    """
    LOGFMT = '[{metric_name}] value: {value}'

    def __init__(self, level):
        self.level = getattr(logging, level.upper())
        self._logger = logging.getLogger('gordon.metrics-logger')

    def log(self, metric):
        """Format and output metric.

        Args:
            metric (dict): Complete metric.
        """
        message = self.LOGFMT.format(**metric)
        if metric['context']:
            message += ' context: {context}'.format(context=metric['context'])
        self._logger.log(self.level, message)


@zope.interface.implementer(interfaces.IMetricRelay)
class LogRelay:
    """Manage metrics that get output to a logger.

    Args:
        config (dict): Required keys:
            level (str): Log level of metrics log messages.
            time_unit (number): (optional) Scale time unit for use with
                LogTimer, for example: 1E9 to send nanoseconds.
    """
    def __init__(self, config):
        level = config.get('log_level', 'info')
        self.time_unit = config.get('time_unit', 1)
        self.logger = LoggerAdapter(level)
        self.counters = collections.defaultdict(int)

    def _create_metric(self, metric_name, value, context, **kwargs):
        context = context or {}

        return {
            'metric_name': metric_name,
            'value': value,
            'context': context
        }

    async def incr(self, metric_name, value=1, context=None, **kwargs):
        """Increase a metric counter by ``value`` and output result.

        Args:
            metric_name (str): Identifier of the metric.
            value (number): (optional) Value with which to increase the metric.
            context (dict): (optional) Additional key-value pairs which further
                describe the metric, for example: {'remote-host': '1.2.3.4'}
        """
        self.counters[metric_name] += value
        await self.set(metric_name, self.counters[metric_name], context,
                       **kwargs)

    def timer(self, metric_name, context=None, **kwargs):
        """Create a LogTimer.

        Args:
            metric_name (str): Identifier of the metric.
            context (dict): (optional) Additional key-value pairs which further
                describe the metric, for example: {'unit': 'milliseconds'}
        """
        metric = self._create_metric(metric_name, None, context, **kwargs)
        return LogTimer(metric, self.logger, self.time_unit)

    async def set(self, metric_name, value, context=None, **kwargs):
        """Output ``metric_name`` with ``value``.

        Args:
            metric_name (str): Identifier of the metric.
            value (number): (optional) Value of the metric.
            context (dict): (optional) Additional key-value pairs which further
                describe the metric, for example: {'app-version': '1.5.4'}
        """
        metric = self._create_metric(metric_name, value, context, **kwargs)
        self.logger.log(metric)

    async def cleanup(self):
        """Not used."""
        pass



# -*- coding: utf-8 -*-
#
# Copyright 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Main module to run the Gordon service.

The service expects a ``gordon.toml`` and/or a ``gordon-user.toml``
file for configuration in the current working directory, or in a
provided root directory.

Any configuration defined in ``gordon-user.toml`` overwrites those in
``gordon.toml``.

Example:

.. code-block:: bash

    $ python gordon/main.py
    $ python gordon/main.py -c /etc/default/
    $ python gordon/main.py --config-root /etc/default/
"""

import asyncio
import logging
import os
import signal

import click
import toml
import ulogger

from gordon import __version__ as version
from gordon import exceptions
from gordon import interfaces
from gordon import plugins_loader
from gordon import router


async def shutdown(sig, loop):
    """Gracefully cancel current tasks when app receives a shutdown signal."""
    logging.info(f'Received exit signal {sig.name}...')
    tasks = [task for task in asyncio.Task.all_tasks() if task is not
             asyncio.tasks.Task.current_task()]

    for task in tasks:
        logging.debug(f'Cancelling task: {task}')
        task.cancel()

    results = await asyncio.gather(*tasks, return_exceptions=True)
    logging.debug(f'Done awaiting cancelled tasks, results: {results}')

    loop.stop()
    logging.info('Shutdown complete.')


def _load_config(root=''):
    conf, error = {}, False
    conf_files = ['gordon.toml', 'gordon-user.toml']
    for conf_file in conf_files:
        try:
            with open(os.path.join(root, conf_file), 'r') as f:
                conf.update(toml.load(f))
        except IOError:
            error = True

    if error and conf == {}:
        raise IOError(f'Cannot load Gordon configuration file from "{root}".')
    return conf


def setup(config_root=''):
    """
    Service configuration and logging setup.

    Configuration defined in ``gordon-user.toml`` will overwrite
    ``gordon.toml``.

    Args:
        config_root (str): Where configuration should load from,
            defaults to current working directory.
    Returns:
        A dict for Gordon service configuration.
    """
    config = _load_config(root=config_root)

    logging_config = config.get('core', {}).get('logging', {})
    log_level = logging_config.get('level', 'INFO').upper()
    log_handlers = logging_config.get('handlers') or ['syslog']

    ulogger.setup_logging(
        progname='gordon', level=log_level, handlers=log_handlers)

    return config


def _log_or_exit_on_exceptions(base_msg, exc, debug):
    log_level_func = logging.warn
    if not debug:
        log_level_func = logging.error

    if isinstance(exc, list):
        for exception in exc:
            log_level_func(base_msg, exc_info=exception)
    else:
        log_level_func(base_msg, exc_info=exc)

    if not debug:
        raise SystemExit(1)


def _gather_plugins_by_type(plugins, debug):
    runnable_plugins = []
    message_handling_plugins = []
    for plugin in plugins:
        # TODO (lynn): these should be switched out for adding
        # the "verify interface implementation" ability. See
        # https://docs.zope.org/zope.interface/verify.html
        if interfaces.IRunnable.providedBy(plugin):
            if not hasattr(plugin, 'run') or \
                    not asyncio.iscoroutinefunction(plugin.run):
                msg = (f'Implemention "{plugin}" of the required '
                       '"IRunnable" interface does not have the '
                       'necessary `run` method.')
                exc = exceptions.InvalidPluginError(msg)
                _log_or_exit_on_exceptions(msg, exc, debug)
                continue
            runnable_plugins.append(plugin)

        if interfaces.IMessageHandler.providedBy(plugin):
            if not hasattr(plugin, 'handle_message') or \
                    not asyncio.iscoroutinefunction(plugin.handle_message):
                msg = (f'Implemention "{plugin}" of the required '
                       '"IMessageHandler" interface does not have the '
                       'necessary `handle_message` method.')
                exc = exceptions.InvalidPluginError(msg)
                _log_or_exit_on_exceptions(msg, exc, debug)
                continue
            message_handling_plugins.append(plugin)

    if not runnable_plugins or not message_handling_plugins:
        msg = (f'At least one runnable plugin is required.')
        exc = exceptions.MissingPluginError(msg)
        _log_or_exit_on_exceptions(msg, [exc], debug=debug)

    return runnable_plugins, message_handling_plugins


def _setup_router(config, plugins, metrics, success_channel, error_channel):
    msg_router = router.GordonRouter(
        config, success_channel, error_channel, plugins, metrics)
    return msg_router


async def _run(runnable_plugins, msg_router, debug):
    tasks = [p.run() for p in runnable_plugins]
    tasks.append(msg_router.run())
    await asyncio.gather(*tasks)


@click.command()
@click.option('-c', '--config-root',
              type=click.Path(exists=True), required=False, default='.',
              help='Directory where to find service configuration.')
def run(config_root):
    config = setup(os.path.abspath(config_root))
    debug_mode = config.get('core', {}).get('debug', False)

    channels = {
        'success_channel': asyncio.Queue(),
        'error_channel': asyncio.Queue(),
    }
    plugin_names, plugins, errors, metrics = plugins_loader.load_plugins(
        config, channels)
    if errors:
        for err_plugin, exc in errors:
            base_msg = 'Plugin was not loaded: {err_plugin}'
            _log_or_exit_on_exceptions(base_msg, exc, debug=debug_mode)

    if plugin_names:
        logging.info(f'Loaded {len(plugin_names)} plugins: {plugin_names}.')

    runnables, message_handlers = _gather_plugins_by_type(plugins, debug_mode)

    route_config = config.get('core', {}).get('route', {})
    msg_router = _setup_router(
        route_config, message_handlers, metrics, **channels)

    logging.info(f'Starting gordon v{version}...')
    loop = asyncio.get_event_loop()

    # Register shutdown to signals

    for signame in (signal.SIGHUP, signal.SIGTERM, signal.SIGINT):
        loop.add_signal_handler(
            signame, lambda: asyncio.ensure_future(shutdown(signame, loop)))

    try:
        loop.create_task(_run(runnables, msg_router, debug_mode))
        loop.run_forever()

    finally:
        loop.close()


if __name__ == '__main__':
    run()



# -*- coding: utf-8 -*-
# Copyright (c) 2018 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numbers

import pytest

from gordon.metrics import ffwd


def test_ffwd_protocol_connection_made(mocker):
    """Message sends and connection closes."""
    mock_transport = mocker.Mock()
    message = 'Testing message!'

    protocol = ffwd.UDPClientProtocol(message)
    protocol.connection_made(mock_transport)

    mock_transport.sendto.assert_called_once_with(message)
    mock_transport.close.assert_called_once_with()


@pytest.mark.asyncio
async def test_udp_client_send(mocker):
    """Sends serialized metric using FfwdClientProtocol."""
    mock_loop = mocker.Mock()

    actual_func = None
    actual_addr = None

    async def mock_create_datagram_endpoint(func, remote_addr):
        nonlocal actual_func
        nonlocal actual_addr
        actual_func = func
        actual_addr = remote_addr

    mock_loop.create_datagram_endpoint = mock_create_datagram_endpoint

    udp_client = ffwd.UDPClient('1.2.3.4', '5678', mock_loop)
    await udp_client.send({'an': 'object'})

    actual_protocol = actual_func()
    assert isinstance(actual_protocol, ffwd.UDPClientProtocol)
    assert ('1.2.3.4', '5678') == actual_addr
    assert b'{"an": "object"}' == actual_protocol.message


@pytest.fixture
def mock_udp_client_and_metrics_sent(mocker):
    metrics_sent = []

    async def mock_send(metric):
        nonlocal metrics_sent
        metrics_sent.append(metric)

    mock_udp_client = mocker.Mock()
    mock_udp_client.send = mock_send
    return mock_udp_client, metrics_sent


@pytest.fixture
def timer_and_metrics_sent(mock_udp_client_and_metrics_sent):
    mock_udp_client, metrics_sent = mock_udp_client_and_metrics_sent
    metric = {'a': 'metric', 'value': None}
    timer = ffwd.FfwdTimer(metric, mock_udp_client)
    return timer, metrics_sent


@pytest.mark.asyncio
async def test_ffwdtimer_manual(timer_and_metrics_sent):
    """Times when operated manually."""
    timer, metrics_sent = timer_and_metrics_sent
    await timer.start()
    assert isinstance(timer._start_time, numbers.Number)

    await timer.stop()
    assert 1 == len(metrics_sent)
    assert 'metric' == metrics_sent[0]['a']
    assert 0 < metrics_sent[0]['value']


@pytest.mark.asyncio
async def test_ffwdtimer_as_context_manager(timer_and_metrics_sent):
    """Times as a context manager."""
    timer, metrics_sent = timer_and_metrics_sent
    async with timer as t:
        assert isinstance(t._start_time, numbers.Number)

    assert 1 == len(metrics_sent)
    assert 'metric' == metrics_sent[0]['a']
    assert 0 < metrics_sent[0]['value']


@pytest.fixture
def test_config():
    return {
        'key': 'service-name',
        'ffwd_ip': '1.2.3.4',
        'ffwd_port': '5678',
        'time_unit': 1E9
    }


TEST_KEY_NAME_VAL = [
    ('service-name', 'metric-name', 10),
    ('other-name', 'metric-name', 6.243289),
    ('service-name', 'other-metric', 500000),
    ('service-name', 'other-metric', -20),
    ('service-name', 'other-metric', 10),
]


TEST_KWARGS = [
    {'context': None, 'tags': None, 'other_extraneous_kwarg': 'ignored!'},
    {'context': {'remote-host': 'dns-host-3'}, 'tags': None},
    {'context': None, 'tags': ['v40']},
    {'context': {'type': 'active'}, 'tags': ['v40']},
    {'context': {'type': 'active'}, 'extraneous_kwarg': 'should be ignored'}
]


@pytest.mark.parametrize('key_name_val,kwargs', zip(
    TEST_KEY_NAME_VAL, TEST_KWARGS))
def test_ffwdrelay_create_metric(test_config, key_name_val, kwargs):
    """Creates metric for all arguments."""
    key, name, val = key_name_val
    test_config['key'] = key
    relay = ffwd.SimpleFfwdRelay(test_config)
    actual = relay._create_metric(name, val, **kwargs)

    attrs = kwargs['context'] or {}
    attrs.update({'what': name})
    expected = {
        'key': key,
        'attributes': attrs,
        'value': val,
        'type': 'metric'
    }
    assert expected == actual


@pytest.fixture
def ffwdrelay_and_metrics_sent(test_config, mock_udp_client_and_metrics_sent):
    mock_udp_client, metrics_sent = mock_udp_client_and_metrics_sent
    relay = ffwd.SimpleFfwdRelay(test_config)
    relay.udp_client = mock_udp_client
    return relay, metrics_sent


@pytest.mark.asyncio
@pytest.mark.parametrize('kwargs', TEST_KWARGS)
async def test_ffwdrelay_incr(ffwdrelay_and_metrics_sent, kwargs):
    """Creates and sends metrics with value 1."""
    ffwdrelay, metrics_sent = ffwdrelay_and_metrics_sent
    expected = [ffwdrelay._create_metric('some-increasing-value', 1, **kwargs)]
    await ffwdrelay.incr('some-increasing-value', **kwargs)
    assert expected == metrics_sent


@pytest.fixture
def ffwdrelay(ffwdrelay_and_metrics_sent):
    return ffwdrelay_and_metrics_sent[0]


@pytest.mark.parametrize('kwargs', TEST_KWARGS)
def test_ffwdrelay_timer(ffwdrelay, kwargs):
    """Returns initialized timer."""
    expected_metric = ffwdrelay._create_metric('some-timer', None, **kwargs)
    actual = ffwdrelay.timer('some-timer', **kwargs)
    assert isinstance(actual, ffwd.FfwdTimer)
    assert expected_metric == actual.metric
    assert ffwdrelay.time_unit == actual.time_unit
    assert ffwdrelay.udp_client == actual.udp_client


@pytest.mark.asyncio
@pytest.mark.parametrize('kwargs', TEST_KWARGS)
async def test_ffwdrelay_set(ffwdrelay_and_metrics_sent, kwargs):
    """Creates and sends metric with arbitrary value."""
    ffwdrelay, metrics_sent = ffwdrelay_and_metrics_sent
    expected = [
        ffwdrelay._create_metric('some-measure', 42, **kwargs)
    ]
    await ffwdrelay.set('some-measure', 42, **kwargs)
    assert expected == metrics_sent


@pytest.mark.asyncio
async def test_cleanup(ffwdrelay):
    """Raises no errors."""
    await ffwdrelay.cleanup()



# -*- coding: utf-8 -*-
# Copyright (c) 2017 Spotify AB
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from zope.interface import Attribute
from zope.interface import Interface


class IEventMessage(Interface):
    """A discrete unit of work for Gordon to process.

    Gordon expects plugins to return or accept objects that implement this
    interface in order to route them to other plugins, and handle retries
    or cleanup in case of errors.
    """

    msg_id = Attribute('Identifier for the event message instance.')
    phase = Attribute('Variable phase of the event message.')

    def __init__(msg_id, data, history_log, phase=None):
        """Initialize an EventMessage.

        Args:
            msg_id (str): Unique message identifier.
            data (dict): Data required to update DNS records.
            history_log (list): Log of actions performed on message.
            phase (str): Current phase.
        """

    def append_to_history(entry, plugin_phase):
        """Append entry to the IEventMessage's history_log.

        Args:
            entry (str): Information to append to log.
            plugin_phase (str): Phase of plugin that created the log
                entry message.
        """

    def update_phase(new_phase):
        """Update the phase of a message to new phase.

        Args:
            new_phase (str): Phase to update the message to.
        """


class IGenericPlugin(Interface):
    """**Do not** implement this interface directly.

    Use :py:class:`gordon.interfaces.IRunnable`,
    or :py:class:`gordon.interfaces.IMessageHandler` instead.
    """

    async def shutdown():
        """Perform any actions required to gracefully shutdown plugin."""


class IRunnable(IGenericPlugin):
    """Runnable plugin to produce event messages for Gordon to process.

    The plugin also has the ability to send :py:class:`gordon.interfaces
    EventMessage` objects to both success and error channels. At least
    one runnable plugin is required to run Gordon.
    """

    start_phase = Attribute('Starting phase for event messages.')

    def __init__(config, success_channel, error_channel, metrics, **kwargs):
        """Initialize a runnable plugin.

        Args:
            config (dict): Plugin-specific configuration.
            success_channel (asyncio.Queue): A sink for successfully processed
                IEventMessages.
            error_channel (asyncio.Queue): A sink for IEventMessages that were
                not processed due to problems.
            metrics (obj): Optional obj used to emit metrics.
        """

    async def run():
        """Begin consuming messages using the provided event loop."""


class IMessageHandler(IGenericPlugin):
    """Plugin which performs some operation on an event message.

    The Gordon core router will use its `phase_route` to direct messages
    produced by any runnable plugins the appropriate message handling
    plugins, identified by their phase attribute. At least
    one message handling plugin is required to run Gordon.
    """

    phase = Attribute('Plugin phase')

    def __init__(config, metrics, **kwargs):
        """Initialize a message handler.

        Args:
            config (dict): Plugin-specific configuration.
            metrics (obj): Obj used to emit metrics.
        """

    async def handle_message(event_message):
        """Perform some operation on or triggered by an event message.

        Args:
            event_message (IEventMessage): Message on which to operate.
        """


class IMetricRelay(Interface):
    """Manage Gordon metrics."""

    async def incr(metric_name, value=1, context=None, **kwargs):
        """Increase a metric by 1 or a given amount.

        Args:
            metric_name (str): Identifier of the metric.
            value (int): (optional) Value with which to increase the metric.
            context (dict): (optional) Additional key-value pairs which further
                describe the metric, for example: {'remote-host': '1.2.3.4'}
        """

    def timer(metric_name, context=None, **kwargs):
        """Get a timer object which implements ITimer.

        Args:
            metric_name (str): Identifier of the metric.
            context (dict): (optional) Additional key-value pairs which further
                describe the metric, for example: {'unit': 'seconds'}
        """

    async def set(metric_name, value, context=None, **kwargs):
        """Set a metric to a given value.

        Args:
            metric_name (str): Identifier of the metric.
            value (number): The value of the metric.
            context (dict): (optional) Additional key-value pairs which further
                describe the metric, for example: {'app-version': '1.5.3'}
        """

    async def cleanup(**kwargs):
        """Perform cleanup tasks related to metrics handling."""


class ITimer(Interface):
    """Timer supporting both manual operation and use as a context manager."""

    async def __aenter__():
        """Enter context manager to start timing."""

    async def __aexit__(type, value, traceback):
        """Exit context manager to stop timing."""

    async def start():
        """Start timing."""

    async def stop():
        """End timing."""
